/*
 *  link.c	- routines to open/close serial link for PPP
 *
 * Copyright 1992-1993 Merit Network, Inc. and The Regents of the
 *  University of Michigan.  Usage of this source code is restricted
 *  to non-profit, non-commercial purposes.  The source is provided
 *  "as-is", without warranty.
 */
#include "ppp.h"
#include <Folders.h>
#include <GestaltEqu.h>
#include <CommResources.h>
#include <CRMSerialDevices.h>

/* Trap numbers */
#define TN_UnknownOS		0xA09F
#define TN_CommToolbox		0xA08B

unsigned char PortBUse : 0x291;

/* function prototypes */
OSErr waitstr(LapInfo *lap, char *, int);
OSErr sendstr(unsigned char *);
void install_tm(LapInfo *, struct TMprocess *, ProcPtr);
void AppendStr(unsigned char *, unsigned char *);
OSErr parsestr(unsigned char *, Boolean);
OSErr TermMode();
void sendlcpecho();
void time_check();
void MySetText(unsigned char);
pascal Boolean nullfilt(DialogPtr, EventRecord *, int *);

/* define some globals */
char **gmessages;
Handle ghandle;
char reserrmes[]="\pMacPPP: resource error";
char okmes[]={1, 2,'O','K',0};
extern struct NMRec errnmrec;
char resFile[] = "\pPPP";
char sys6resFile[] = "\p:System Folder:PPP";

/* resource defines */
#define SYS6FOLDER -4034
#define PREFFILENAME -4033

#define MODEMRESPONSES 128 /* resource id of modem responses */
#define STATUSSTRINGS 129
#define MODEMCOMMANDS 130
#define PHASETEXT 131

#define MODEMCONNECT 1	/* # for connect message */
#define MODEMBUSY 2		/* # for busy message */

#define MESSCHECK 1
#define MESSINIT 2
#define MESSDIAL 3
#define MESSBUSY 4
#define MESSERROR 5
#define MESSSENDING 6
#define MESSWAITING 7
#define MESSTERM 8
#define MESSTIMEOUT 9
#define MESSPHASE 10

#define MODEMALERT	128		/* Modem connect failure alert */
#define NOPREFALERT	129		/* no preferences alert */
#define SERIALALERT 130		/* alert for error when opening serial port */
#define TIMEOUTALERT 131	/* timeout waiting for a string */
#define IDLEALERT	132
#define ECHOALERT	133
#define PAPMESALERT 134
#define OLDPREFSALERT 135
#define INUSEALERT	136		/* port in use alert */

/* dialogs */
#define PPPSTATUS	129		/* status dialog */
#define TERMDLOG	130		/* terminal window dialog box */

#define IGNOREIDLE 2
#define IGNOREECHO	2
#define CLOSEPPP	3

void
link_open(register LapInfo *lap)
{
register OSErr		rc, savrc;
OSErr		waitresult;
SerShk		hshake;
Boolean		HaveFindFolder = false, HavePrefFolder = false;
long		ffresp, count, whichDir, oldA4, oldA5, timeout;
Byte		i;
int			whichVol, rn = -1, len, prefref;
int			savresfile, itemhit, type;
char		*strptr;
register DialogPtr	statusdlog;
DialogPtr	dlog;
Rect		rect;
Handle		itemH, itemH2;
pascal		Boolean modemfilt();
struct		bufheader *bufptr;
b_8			curphase, curstate;
char		**modemstrings;
char		tempstr[MAXSLEN + 4];
char		indriv[128],outdriv[128];
long		qd_save;

	/* save current Resource file */
	savresfile = CurResFile();
	
	oldA4 = seta4((long)lap->LapA4);	/* setup PPP's A4 for globals */
	oldA5 = seta5((long) CurrentA5);
	qd_save = *((long *)geta5());	/* save current pointer to QD globals */
	InitGraf(&thePort);		/* init our QD globals */
	InitFonts();			/* init fonts */
	GetWMgrPort(&dlog);		/* check for window manager port */
	if (dlog == nil) {	/* hopefully, these should be set up, but maybe not */
		InitWindows();
		if (TEScrpHandle ==  nil)
			TEInit();
		InitDialogs(nil);
	}
	InitCursor();	/* make the cursor an arrow */

	/* see if findfolder is available */
	if ( Gestalt(gestaltFindFolderAttr, &ffresp) == noErr) {
		if ( ( 1L << gestaltFindFolderPresent ) & ffresp )
			HaveFindFolder = true;
	}
	
	if ( HaveFindFolder ) {
		if ( FindFolder(kOnSystemDisk,kExtensionFolderType,
				kDontCreateFolder,&whichVol,&whichDir) == noErr)
			rn = HOpenResFile(whichVol,whichDir,resFile,fsRdPerm);
	} else {
		GetVRefNum(SysMap, &whichVol);	/* get system Vol ref */
		rn = OpenRFPerm(sys6resFile,whichVol,fsRdPerm);
	}
	
	if (rn == -1) {
		errnmrec.nmStr = reserrmes;	/* initialize pointer to error message */
		NMInstall(&errnmrec);		/* Install the notification */
		goto noresbye;
	}

	if (lap->ppp_flags & ECHO_FAIL) {
		itemhit = NoteAlert(ECHOALERT, nil);
		if (itemhit == IGNOREECHO) {
			lap->echo_count = 0;
			PrimeTime(&(lap->echo_task), (long)lap->prefdata.echo * 1000L);
			goto exitout;
		}
		if (itemhit == CLOSEPPP)
			goto close_ppp;
		link_close(lap);		/* must want to restart */
	}
		
	if (lap->ppp_flags & IDLE_TIMEOUT) {
		if (!lap->prefdata.quiet) {
			itemhit = NoteAlert(IDLEALERT, nil);
			if (itemhit == IGNOREIDLE) {
 				lap->idle_timer = 0;
				PrimeTime(&(lap->timeout_task), 60000L); /* one minute interval */
				goto exitout;
			}
		}
		goto close_ppp;
	}
	
	lap->ppp_flags &= ~(ECHO_FAIL | IDLE_TIMEOUT);

	if (lap->ppp_phase != pppDEAD)
		return;		/* return if not in DEAD phase */

	if (HaveFindFolder) {
		if (noErr == FindFolder(kOnSystemDisk,kPreferencesFolderType,
				kCreateFolder,&whichVol,&whichDir))
			HavePrefFolder = true;
	}

	tempstr[0] = 0;
	if (!HavePrefFolder) {
		itemH = GetResource('STR ',SYS6FOLDER); /* system 6 folder */
		AppendStr(tempstr, *((unsigned char **) itemH));
		ReleaseResource(itemH);
	}
	itemH = GetResource('STR ',PREFFILENAME); /* get pref file */
	AppendStr(tempstr, *((unsigned char **) itemH));
	ReleaseResource(itemH);

	if (HavePrefFolder)
		rc = HOpen(whichVol,whichDir,tempstr,fsRdPerm,&prefref);
	else {
	/* if we don't have PrefFolder, just use System Folder */
		GetVRefNum(SysMap, &whichVol);	/* get system Vol ref */
		rc = FSOpen(tempstr,whichVol,&prefref);
	}
	savrc = rc;

	count = sizeof (struct ppp_pref);
	if (rc == opWrErr )  /* must already be opened by control panel */
		rc = SetFPos(prefref,fsFromStart,0L);
	if (rc == noErr)
		rc = FSRead(prefref, &count, &(lap->prefdata)); /* read in preferences data */

	if (lap->prefdata.version < PREF_VERSION) {
		NoteAlert(OLDPREFSALERT,nil);
		FSClose(prefref);
		goto exitout;
	}
	if (rc == noErr) {
		if	(count != sizeof (struct ppp_pref))  /* check number of bytes */
			rc = -1;	/* set return code to non-zero */
	}
	if (rc == noErr)
		rc = SetFPos(prefref, fsFromMark,
				lap->prefdata.active_config * sizeof(struct ppp_config));
	if (rc == noErr) {
			count = sizeof(struct ppp_config);
			rc = FSRead(prefref, &count, &(lap->configdata));
	}
	if (savrc != opWrErr )
		FSClose(prefref);		/* close the preferences file */

	if ( rc != noErr ) {
		NoteAlert(NOPREFALERT, nil);
		goto exitout;
	}
	indriv[0] = 0;
	outdriv[0] = 0;
	if (NGetTrapAddress(TN_CommToolbox, OSTrap)
			!= GetTrapAddress(TN_UnknownOS)) {
		CRMRec *crmrecptr, acrmrec;
		CRMSerialRecord *serrec;
		long old = 0;
		int index = 1;

		InitCRM();
		acrmrec.qType = crmType;
		while (true) {
			acrmrec.crmDeviceType = crmSerialDevice;
			acrmrec.crmDeviceID = old;
			if (!(crmrecptr = (CRMRec *)CRMSearch((QElemPtr)(&acrmrec))))
				break;
			serrec = (CRMSerialRecord *)crmrecptr->crmAttributes;
			old = crmrecptr->crmDeviceID;
			++index;
			if (EqualString(lap->prefdata.portname, *serrec->name,
					FALSE, TRUE))
				break;
		}
		if (!crmrecptr)
			goto sererr;
		AppendStr(indriv, *serrec->inputDriverName);
		AppendStr(outdriv, *serrec->outputDriverName);

	} else {
		if (!(itemH = GetNamedResource('Port',lap->prefdata.portname)))
			goto sererr;
		strptr = * ((char **) itemH);
		AppendStr(indriv,strptr);
		strptr += *strptr + 1;
		AppendStr(outdriv,strptr);
		ReleaseResource(itemH);
	}
	rc = OpenDriver(indriv, &lap->serinrefnum);
	if (rc == portInUse) {
		if (EqualString(indriv,"\p.BIn", FALSE, FALSE)) {
			if (NoteAlert(INUSEALERT, nil) == OK) {
				PortBUse = (Byte) 0xff;		/* Hack to mark port as free */
				rc = OpenDriver(indriv, &lap->serinrefnum);
			}
		}
	}
	if (rc == noErr) {
		rc = OpenDriver(outdriv, &lap->seroutrefnum);
		if (rc != noErr)
			CloseDriver(lap->serinrefnum);
	}
	
	if (rc != noErr)
		goto sererr;
	rc = SerReset(lap->serinrefnum,stop10+noParity+data8);
	if (rc == noErr)
		rc = SerReset(lap->seroutrefnum,stop10+noParity+data8);
	if (rc == noErr)
		rc = SerSetBuf(lap->serinrefnum, lap->sdinbuf, SDINBUFLEN);	/* give driver a new buffer */

	lap->stat_pb.csCode = 13;		/* baud rate routine */
	*( (unsigned int *) lap->stat_pb.csParam) = lap->configdata.baudrate;
	lap->stat_pb.ioCRefNum = lap->serinrefnum;
	asm {
		lea	lap->stat_pb,a0
		_PBControl	IMMED
	}
	*( (unsigned int *) lap->stat_pb.csParam) = lap->configdata.baudrate;
	lap->stat_pb.ioCRefNum = lap->seroutrefnum;
	asm {
		lea	lap->stat_pb,a0
		_PBControl	IMMED
	}

	hshake.fDTR = false; 	/* don't bother with input flow ctl */
	hshake.fCTS = lap->configdata.flowctl_on; /* enable/disable flow ctl */
	hshake.errs = 0;	/* don't abort reads on errors */
	hshake.evts = 0;	/* don't interrupt on events */
	hshake.fXOn = false; /* no Xon/Xoff flow control */
	hshake.fInX = false;
	
	if (rc == noErr)
		rc = SerHShake(lap->serinrefnum,&hshake);
	if (rc == noErr)
		rc = SerHShake(lap->seroutrefnum,&hshake);
	if (rc != noErr) {
		SerSetBuf(lap->serinrefnum, lap->sdinbuf, 0); /* reset buffer */
		CloseDriver(lap->serinrefnum);
		CloseDriver(lap->seroutrefnum);
sererr:
		NoteAlert(SERIALALERT, nil);
		goto exitout;
	}
	
	/* Initialize lists of buffers */
	lap->buflist = (struct bufheader * ) nil;
	for ( i = 0; i < NUMBUFFERS; i++ ) {
		bufptr = lap->blockarray[i].block;
		bufptr->dataptr = (unsigned char *) lap->buflist;
		lap->buflist = bufptr;
	}
	
	lap->rds.rdsparm.lnb.lnb_ptr = nil;		/* clear rcv buffer pointer */

	lap->active = nil;
	lap->out_q.qFlags = 0;
	/* initialize queue of iopbs */
	lap->pppbq.qFlags = 0;
	lap->pppbq.qHead = nil;
	lap->pppbq.qTail = nil;
	for (i = 0; i < NUMIOPBS; i++)
		Enqueue(&(lap->pppiopbs[i]), &(lap->pppbq));

	/* initialize FIFO params */
	lap->XmitQHead = 0;
	lap->XmitQTail = 0;
	lap->XmitQSize = 0;
	
	/* initialize our serial driver iopb's */
	lap->w_iopb.lap = lap;
	lap->w_iopb.iop.ioCompletion = hdlcwioc;	/* where to go when done */
	lap->w_iopb.iop.ioRefNum = lap->seroutrefnum;	/* set output to serial port*/
	
	lap->r_iopb.lap = lap;
	lap->r_iopb.iop.ioRefNum = lap->serinrefnum;
	lap->r_iopb.iop.ioBuffer = lap->rxbuf;
	lap->r_iopb.iop.ioCompletion = SerReadDone;

	lap->stat_pb.ioCRefNum = lap->serinrefnum;
	
	lap->read_state = s_Init;
	lap->write_state = s_Init;
	lap->term_mode = false;	/* not in terminal emulation mode */
	
	lap->bufptr = getbuffer();	/* allocate initial rcv. buffer */
	lap->rddata = lap->bufptr->dataptr;	/* point to data area */
	lap->ok_to_xmit = true;

	SystemTask();		/* some serial drivers need SystemTask time to open */
	Delay(60L, &count);	/* wait some (if DTR need to "wake up" device */
	SystemTask();		/* more system task time for good measure */

	/* set up receive time manager task */
	install_tm(lap, &(lap->rxp_task.atm), ProcRcvPPP);
	PrimeTime(&(lap->rxp_task), 3L);	/* Prime receive task */

	/* set up transmit time manager task (gets primed as needed) */
	install_tm(lap, &(lap->txp_task.atm), ProcXmtPPP);
	
	if (lap->prefdata.echo != 0)	/* LCP echo test */
 		install_tm(lap, &(lap->echo_task.atm), sendlcpecho);
 	if (lap->prefdata.timeout != 0)	/* idle timeout */
 		install_tm(lap, &(lap->timeout_task.atm), time_check);

	statusdlog = GetNewDialog(PPPSTATUS, nil, (WindowPtr) -1L);
	DrawDialog(statusdlog);
	SetCursor(&qd.arrow);
	GetDItem(statusdlog, 3, &type, &ghandle, &rect);
	GetDItem(statusdlog, 4, &type, &itemH2, &rect);
	SetIText(itemH2, "\p");	/* set to nil */
	gmessages = Get1Resource('STR#', STATUSSTRINGS);
	
	if (lap->prefdata.use_term) {
		if (TermMode() != noErr)
			goto getout;
		goto start_ppp;
	}
	
	if (lap->configdata.phonenum[0] != 0) {
retrymodem:
		DrawDialog(statusdlog);
		MySetText(MESSCHECK);
		sendstr("\pATE0V1\r");		/* see if modem is there */
		if ( (rc = waitstr(lap, okmes, 5))
					== 0) { /* wait for response */
			goto retrymodem;
		} else if ( rc == -1 )
			goto getout;

		if ( i = lap->configdata.modeminit[0] ) {
			MySetText(MESSINIT);
			lap->configdata.modeminit[0] = 2;
			if (! EqualString(lap->configdata.modeminit, "\pAT", FALSE, FALSE) )
				sendstr("\pAT");
			lap->configdata.modeminit[0] = i;
			sendstr(lap->configdata.modeminit); /* send modem init command */
			sendstr("\p\r");		/* add a carriage return */
			if ((rc = waitstr(lap, okmes, 5)) == 0) { /* wait for response */
				goto retrymodem;
			} else if (rc == -1)
				goto getout;
		}

		MySetText(MESSDIAL);
doredial:
		modemstrings = Get1Resource('STR#',MODEMRESPONSES);
		DrawDialog(statusdlog);
		if (lap->prefdata.use_pulse)
			sendstr("\pATDP");
		else
			sendstr("\pATDT");
		sendstr(lap->configdata.phonenum);
		sendstr("\p\r");
		HLock((Handle) modemstrings);
		rc = waitstr(lap, *modemstrings + 1,
						lap->configdata.connecttimeout);
		HUnlock((Handle) modemstrings);
		ReleaseResource(modemstrings);
		if (rc == MODEMBUSY) {
			MySetText(MESSBUSY);
			sendstr("\pAT\r");	/* get the modems' attention */
			Delay(120L, &count);			/* add some delay before we redial */
			goto doredial;
		}
		if (rc == 0)
			goto retrymodem;
		if (rc == -1)
			goto getout;
		if (rc != MODEMCONNECT) {
			ResetAlrtStage();
			if ( NoteAlert(MODEMALERT, nil) == 1 ) {	/* alert user */
				goto getout;
			}
			goto doredial;
		}
	}

retrystr:
	DrawDialog(statusdlog);
	for ( i=0 ; i < NUMCOMMANDS; i++) {
		strptr = tempstr;
		*strptr++ = 1;
		*strptr = 0;
		AppendStr(strptr, lap->configdata.commands[i].scriptstr);
		if (lap->configdata.commands[i].addreturn)
			AppendStr(strptr, "\p\\r");
		if ( *strptr != 0 ) {
			if ( lap->configdata.commands[i].sendout ) {
				MySetText(MESSSENDING);
				SetIText(itemH2, strptr);
				parsestr(strptr, true);
			} else {
				MySetText(MESSWAITING);
				SetIText(itemH2, strptr);
				parsestr(strptr, false);
				if ((rc = waitstr(lap, tempstr,
						lap->configdata.waittimeout)) == 0)
					goto retrystr;
				else if ( rc == -1 )
					goto getout;
			}
		}
	}

start_ppp:
	
	lap->write_state = s_Idle;		/* serial port now ready for PPP */
	lap->read_state = s_Idle;

	ppp_iostatus(lap, PARAM_UP);	/* let PPP know that link is up */
	
	curphase = 255;		/* initialize state holder */

	MySetText(MESSPHASE);
	ReleaseResource(gmessages);	/* release messages resource */
	gmessages = Get1Resource('STR#', PHASETEXT);
	ghandle = itemH2;
	while (lap->ppp_fsm[IPcp].state != fsmOPENED) {	/* wait for IPCP */
		ModalDialog(nullfilt, &itemhit);
		if (itemhit == 1) {		/* check if user wants to quit */
			goto getout;
		}
		if (lap->ppp_phase != curphase) {
			curphase = lap->ppp_phase;
			MySetText(curphase + 1);	/* display the PPP phase */
		}
		if (lap->pap_i.IOFlag == TRUE && lap->pap_i.message[0] == 0)
			if (pap_userio(&(lap->ppp_fsm[Pap])) != noErr ) { /* get user name and password */
				goto getout;
			}
		if (lap->pap_i.message[0] != 0) {
			if (!lap->prefdata.quiet || lap->pap_i.IOFlag == true) {
				ResetAlrtStage();
				ParamText(lap->pap_i.message, nil, nil, nil);
				Alert(PAPMESALERT, nil);	/* display PAP messages */
			}
			lap->pap_i.message[0] = 0;	/* reset the message */
		}
	}

	DisposDialog(statusdlog);
 	if (lap->prefdata.echo != 0) {
 		lap->echo_count = 0;
		PrimeTime(&(lap->echo_task), (long)lap->prefdata.echo * 1000L);
	}

 	if (lap->prefdata.timeout != 0) {
 		lap->idle_timer = 0;
		PrimeTime(&(lap->timeout_task), 60000L); /* one minute interval */
	}
	goto exitout;		/* everything okey-dokey */

getout:
	DisposDialog(statusdlog);	/* get rid of status dialog */
	ReleaseResource(gmessages);	/* release resource */
close_ppp:
	PPPClose(lap);		/* set fsm state to INITIAL for LCP and IPCP */
exitout:
	CloseResFile(rn);	/* make sure to close the resource file */
noresbye:
	*((long *)geta5()) = qd_save;	/* restore QD globals ptr */
	seta4(oldA4);		/* restore previous value of A4 */
	seta5(oldA5);
	UseResFile(savresfile);	/* restore resource file */
}

void
link_close(register LapInfo *lap)
{
	int		i;
	long	count;

	ppp_iostatus(lap, PARAM_DOWN);	/* link is going down */
	if (lap->transProc != nil)
		(*(lap->transProc))(TransitionClose);	/* let MacTCP know we are down */

 	if (lap->prefdata.echo != 0)
		RmvTime(&(lap->echo_task));
 	if (lap->prefdata.timeout != 0)
		RmvTime(&(lap->timeout_task));
	RmvTime(&(lap->rxp_task));
	KillIO(lap->serinrefnum);
	RmvTime(&(lap->txp_task));
	KillIO(lap->seroutrefnum);		/* kill pending writes */
	SerSetBuf(lap->serinrefnum, lap->sdinbuf, 0); /* reset buffer */

	if (lap->prefdata.hangup) {	/* hangup modem */
		Delay(65L, &count);		/* wait ~1 second */
		for (i = 0 ; i < 3 ; i++) {
			sendstr("\p+");		/* send a '+' 3 times */
			Delay(12L, &count);	/* wait 200 milliseconds */
		}
		Delay(65L, &count);		/* wait ~1 second */
		sendstr("\pATH\r");		/* send the hang-up command */
		Delay(30L, &count);		/* wait another 1/2 second */
	}

	CloseDriver(lap->serinrefnum);
	CloseDriver(lap->seroutrefnum);
}

void
MySetText(unsigned char stringnum)
{
register unsigned char *strptr;

	strptr = *gmessages;
	strptr++;
	if (stringnum > *strptr++ )		/* check if string is out of range */
		return;
	while (--stringnum > 0)			/* find string in list */
		strptr += *strptr + 1;
	HLock((Handle) gmessages);		/* lock down the Handle */
	SetIText(ghandle, strptr);	/* set the text in the dialog */
	HUnlock((Handle) gmessages);
}

pascal Boolean modemfilt(DialogPtr dlog, EventRecord *evt, int *item)
{
char		c;
ioParam		iopb;
LapInfo		*lap;

	lap = GetLAPPtr();
	*item = 0;
	if (evt->what == keyDown) {
		c = evt->message & charCodeMask;
		iopb.ioRefNum = lap->seroutrefnum;
		iopb.ioBuffer = &c;		/* set data pointer */
		iopb.ioReqCount = 1;		/* number of bytes to send */
		PBWrite(&iopb, false);			
		*item = 3;
	} else if (evt->what == mouseDown)
		return FALSE;
		
	return TRUE;
}

pascal Boolean nullfilt(DialogPtr dlog, EventRecord *evt, int *item)
{
	*item = 0;
	if (evt->what == mouseDown)
		return false;
	return true;
}

OSErr parsestr(unsigned char *instr, Boolean sendout)
{
OSErr			rc;
long			count;
register char	*outptr;
register char	*inptr;
unsigned char	inlen;
register char	c;

	outptr = inptr = instr;
	if ( (inlen = (*outptr++ = *inptr++) ) == 0 )
		return;
	while ( inlen-- != 0 ) {
		if ( ( c = *inptr++ ) == '^' ) {
			if ( inlen-- == 0)		/*  and check for 0 */
				break;
			c = *inptr++ & 0x1F;	/* generate the control char */
			*outptr++ = c;
		} else if (c == '\\') {
			c = *inptr++;
			if ( inlen-- == 0)
				break;
			switch (c) {
			case '^':			/* literals */
			case '\\':
				*outptr++ = c;
				break;
			case 'r':
				*outptr++ = 13;		/* carriage return */
				break;
			case 'b':
			case 'd':
			case 't':
				if (!sendout)		/* only allow on outgoing strings */
					break;
				*instr = (unsigned char) (outptr - instr - 1);
				sendstr(instr);		/* send the current string */
				instr = outptr - 1;	/* point to current place in string */
				if ( c == 'd' )
					Delay(60L, &count);		/* Delay 1 second */
				else if (c == 'b') {
					LapInfo *lap;
					
					lap = GetLAPPtr();		/* get our Lap Pointer */
					SerSetBrk(lap->seroutrefnum);	/* put line in breaking state */
					Delay(6L, &count);				/* wait a huner' milliseconds */
					SerClrBrk(lap->seroutrefnum);	/* clear break condition */
				} else {
					if ( (rc = TermMode()) != noErr)
						return rc;
				}
				break;
			default:
				if  ( c >= '0' && c <= '7' ) {
					*outptr = 0;
					count = 3;
					do {
						*outptr <<= 3;
						*outptr += c - '0';
						c = *inptr++;
					} while ( c >= '0' && c <= '7' && --count != 0 && inlen-- != 0);
					inptr--;	/* point back to charcter */
					outptr++;
					if (inlen == 0xff)
						inlen = 0;
				}
				break;
			}
		} else {
			*outptr++ = c;
		}
	}
	*instr = (unsigned char) (outptr - instr - 1); /* store len byte */
	if (sendout)
		sendstr(instr);		/* send the string */
}

 /* simple terminal window routine */
 
OSErr TermMode()
{
LapInfo 	*lap;
TEHandle	teH;
Rect		rect;
register DialogPtr	dlog;
Handle		itemH;
int			type, len, nlines, itemhit;
GrafPtr		oldport;
	
	lap = GetLAPPtr();
	lap->term_mode = true;		/* in terminal emulation mode */
	MySetText(MESSTERM);
	dlog = GetNewDialog(TERMDLOG, nil, (WindowPtr) -1L);
	GetPort(&oldport);
	SetPort(dlog);
	DrawDialog(dlog);	/* draw the dialog so title shows up */

	GetDItem(dlog, 3, &type, &itemH, &rect);
	TextFont(courier);
	TextSize(10);

	teH = TENew(&rect, &rect);	
	len = 0;
	nlines = 10;	
	FrameRect(&rect);

	do {
		ModalDialog(modemfilt, &itemhit);

		if (lap->bufptr->length != len) {
			len = lap->bufptr->length;
			EraseRect(&rect);
			TESetText(lap->bufptr->dataptr, (long)len, teH);
			TECalText(teH);
			if ((**teH).nLines > nlines) {
				TEScroll(0, (**teH).lineHeight * -1 * ((**teH).nLines - nlines), teH);
				nlines = (**teH).nLines;
			}
			TEUpdate(&rect, teH);
			FrameRect(&rect);
		}
	} while (itemhit != OK && itemhit != Cancel);
		
	lap->term_mode = false;
	DisposeDialog(dlog);	/* Dispose the Dialog */
	SetPort(oldport);
	if (itemhit == Cancel)
		return -1;
	return noErr;
}

/* routine to send out a string.  may want to add timeout code. */

OSErr sendstr(unsigned char *s)
{
ioParam iopb;
LapInfo	*lap;

	if (*s == 0)
		return;				/* check for null string */
	lap = GetLAPPtr();
	iopb.ioRefNum = lap->seroutrefnum;
	iopb.ioReqCount = *s++;	/* get length */
	iopb.ioBuffer = s;		/* set buffer pointer */
	return (PBWrite(&iopb,false));	/* call synchronously */
}

void 
adjustbuf(LapInfo *lap, int i)
{
long	savsr;
struct	bufheader *bufptr = lap->bufptr;
	
	savsr = set_sr(0x2100);	/* disable timer ints, so we don't buffer more */
	bufptr->length -= i;		/* subtract from count */
	lap->rddata -= i;		/* adjust pointer */
	BlockMove(bufptr->dataptr + i, bufptr->dataptr, bufptr->length);	/*move down buffer */
	set_sr(savsr);		/* restore timer interrupts */
}

OSErr waitstr(LapInfo *lap, char *s, int waitseconds )
{
long	timeout;
register int	i;
int		itemhit;
unsigned char	len, numstrs, j;
char	*lenptr;

	timeout = TickCount() + waitseconds*60;	/* seconds before timing out */
	
	for (i=1; true; i++) {
		ModalDialog(nullfilt, &itemhit);
		if (itemhit == 1) {		/* check if user wants to quit */
			return -1;
		}
					
		if (TickCount() > timeout) {		/* don't sit in for loop forever */
			ResetAlrtStage();
			ParamText(s + 1,nil,nil,nil);
			if ( NoteAlert(TIMEOUTALERT, nil) == 1 ) {	/* alert user */
				return -1;
			}
			return 0;				/* user would like to retry */
		}
			
		if (lap->bufptr->length < i) {	/* check if any chars came in */
			--i;
			continue;
		}
		
		if (i > 255) {		/* don't let buffer get too full */
			adjustbuf(lap, 128);
			i -= 128;
		}

		lenptr = s;		/* point to number of string field */
		numstrs = *lenptr++;	/* get the number of strings */
		j = 0;
		while ( j < numstrs ) {
			j++;
			if ((len = *lenptr++) == 0 )
				continue;
			if (i >= len) {			/* do we have enough to do a compare? */
				if ( bytecmp(lap->bufptr->dataptr + i - len, lenptr, len)
						 == 0) {	/* compare and exit if equal */
					adjustbuf(lap, i);	/* shift the data in the buffer */
					return j;			/* return success */
				}
			}
			lenptr += len;		/* point to next string */
		}
	}
}

void sendlcpecho() {
	register LapInfo	*lap;
	struct TMprocess	*tmprocp;

	asm {
		move.l	a1,tmprocp
	}

	lap = tmprocp->tmsavptr.lap;
	if (lap->echo_count++ >= lap->prefdata.echo_tries) {
		if (!(lap->ppp_flags & IDLE_TIMEOUT) && lap->transProc != nil) {
			lap->ppp_flags |= ECHO_FAIL;
			(*(lap->transProc))(TransitionClose);	
			(*(lap->transProc))(TransitionOpen);	/* a hack to get system time */
			return;
		} else
			lap->echo_count = 0;
	}
	 /* send a LCP echo request */
	lap->lcp_echo_buf.header.dataptr = lap->lcp_echo_buf.buffer
			+ LCP_ECHO_BUFSIZE;		/* initialize dataptr */
	lap->lcp_echo_buf.header.length = 0;	/* initialize count */
	fsm_send(&(lap->ppp_fsm[Lcp]), ECHO_REQ, 0, &(lap->lcp_echo_buf));
}

void time_check() {
	register LapInfo	*lap;
	struct TMprocess	*tmprocp;

	asm {
		move.l	a1,tmprocp
	}

	lap = tmprocp->tmsavptr.lap;
	if (lap->idle_timer++ >= lap->prefdata.timeout) {
		if (!(lap->ppp_flags & ECHO_FAIL) && lap->transProc != nil) {
			lap->ppp_flags |= IDLE_TIMEOUT;
			(*(lap->transProc))(TransitionClose);	
			(*(lap->transProc))(TransitionOpen);	/* a hack to get system time */
		} else
			lap->idle_timer = 0;		/* reset idle timer */
	}
	PrimeTime(&(lap->timeout_task), 60000L);	/* re-prime for 1 minute */
}

void install_tm (LapInfo *lap, register struct TMprocess *tmptr, ProcPtr task) {
	bzero((char *) tmptr, (short) sizeof(TMTask));
	tmptr->atm.tmAddr = task;
	tmptr->tmsavptr.lap = lap;
	InsTime(tmptr);
}